#!/usr/bin/env python
#Author: Mark Feineigle
#Create Date: 2018/02/04
#Last Modify: 2018/02/04
# Chinese Remainder Theorem
# (2017)_Stallings_-_Cryptography_and_network_security_principles_and_practice.pdf
# page 72

import sympy as sp

def crt(m1, m2, M, A, verbose=False, solve=True, add=False, mul=False, x=0):
  M1 = M/m1
  M2 = M/m2
  M1inv = sp.mod_inverse(M1, m1)
  M2inv = sp.mod_inverse(M2, m2)

  tA = (A % m1, A % m2) #represent A as tuple
  if solve and not (add or mul): return "tA", tA
  if verbose: print "tA", tA

  # if we want to add or mul x to A
  tx = (x % m1, x % m2) #represent x as tuple
  if verbose and (add or mul): print "tx", tx

  # add the tuples element-wise and mod
  if add:
    tA_plus_x = ((tA[0] + tx[0]) % m1 , (tA[1] + tx[1]) % m2)
    if verbose: print "tA_plus_x", tA_plus_x
    if solve: return (tA_plus_x[0]*M1*M1inv + tA_plus_x[1]*M2*M2inv) % M

  # multiply the tuples element-wise and mod
  if mul:
    tA_mul_x = ((tA[0] * tx[0]) % m1 , (tA[1] * tx[1]) % m2)
    if verbose: print "tA_mul_x", tA_mul_x
    if solve: return (tA_mul_x[0]*M1*M1inv + tA_mul_x[1]*M2*M2inv) % M

def printHeader(m1, m2, M, A):
  print "="*88
  print """m1={} \t\t m2={} \t\t M={} \t\t A={}
tA=tuple A \t tx=tuple x \t tA_plus_x=tuple A+x \t tA_mul_x=tuple A*x""".format(m1,m2,M,A)
  print

if __name__ == "__main__":
  m1 = 37
  m2 = 49
  M = 1813
  A = 973
  printHeader(m1, m2, M, A)
  print "crt(m1=m1, m2=m2, M=M, A=A)"
  print crt(m1=m1, m2=m2, M=M, A=A)
  print
  print "crt(m1=m1, m2=m2, M=M, A=A, add=True, x=678)"
  print crt(m1=m1, m2=m2, M=M, A=A, add=True, x=678)
  print
  print "crt(m1=m1, m2=m2, M=M, A=A, add=True, x=678, verbose=True)"
  print crt(m1=m1, m2=m2, M=M, A=A, verbose=True, add=True, x=678)
  print
  m1 = 37
  m2 = 49
  M = 1813
  A = 1651
  printHeader(m1, m2, M, A)
  print "crt(m1=m1, m2=m2, M=M, A=A, mul=True, x=73)"
  print crt(m1=m1, m2=m2, M=M, A=A, mul=True, x=73)
  print
  print "crt(m1=m1, m2=m2, M=M, A=A, mul=True, x=73, verbose=True)"
  print crt(m1=m1, m2=m2, M=M, A=A, verbose=True, mul=True, x=73)
